import React, { Component } from 'react'
import PropTypes from 'prop-types'
import { is, fromJS } from 'immutable'
import { DndProvider, DragSource, DropTarget } from 'react-dnd'
import { Table, Popover, Modal, message, Button, Typography } from 'antd'
import { PlusOutlined, PlusSquareOutlined, FileSyncOutlined, EditOutlined, CopyOutlined, DeleteOutlined, FontColorsOutlined, CloseCircleOutlined, AntDesignOutlined } from '@ant-design/icons'

import asyncComponent from '@/utils/asyncComponent'
import asyncIconComponent from '@/utils/asyncIconComponent'
import Utils from '@/utils/utils.js'
import MKEmitter from '@/utils/events.js'
import './index.scss'

const { confirm } = Modal
const { Paragraph } = Typography
const EditColumn = asyncComponent(() => import('./editColumn'))
const TableVerify = asyncComponent(() => import('./tableIn'))
const MarkColumn = asyncIconComponent(() => import('@/menu/components/share/markcomponent'))
const ColsControl = asyncIconComponent(() => import('@/menu/components/share/colsControl'))
const CardCellComponent = asyncComponent(() => import('@/menu/components/card/cardcellcomponent'))
const PasteComponent = asyncIconComponent(() => import('@/components/paste'))

class HeaderCol extends Component {
  deleteCol = () => {
    const _this = this

    confirm({
      content: '确定删除显示列吗？',
      onOk() {
        _this.props.deleteCol(_this.props.column)
      },
      onCancel() {}
    })
  }

  updateMarks = (vals) => {
    const { column } = this.props
    this.props.updateCol({...column, marks: vals})
  }

  copycolumn = () => {
    const { column } = this.props

    let oInput = document.createElement('input')
    let val = {
      copyType: 'editcols',
      cols: [column]
    }

    let srcid = localStorage.getItem(window.location.href.split('#')[0] + 'srcId')
    if (srcid) {
      val.$srcId = srcid
    }

    oInput.value = window.btoa(window.encodeURIComponent(JSON.stringify(val)))
    document.body.appendChild(oInput)
    oInput.select()
    document.execCommand('Copy')
    oInput.className = 'oInput'
    oInput.style.display = 'none'

    message.success('复制成功。')

    document.body.removeChild(oInput)
  }

  shouldComponentUpdate (nextProps, nextState) {

    if (this.props.rowSpan !== nextProps.rowSpan || this.props.colSpan !== nextProps.colSpan) {
      return true
    }

    if (!nextProps.column) return true

    return !is(fromJS(this.props.column), fromJS(nextProps.column)) ||
      !is(fromJS(this.props.fields), fromJS(nextProps.fields)) ||
      this.props.index !== nextProps.index ||
      (nextProps.className && this.props.className !== nextProps.className) ||
      window.GLOB.columnId === nextProps.column.uuid ||
      window.GLOB.precolumnId === nextProps.column.uuid
  }

  render() {
    const { connectDragSource, connectDropTarget, moveCol, componentId, dropCol, addElement, updateCol, editColumn, pasteCell, changeStyle, deleteCol, index, column, align, fields, children, ...restProps } = this.props

    if (!column) return (
      <th {...restProps}>{children}</th>
    )

    let style = {cursor: 'move', textAlign: align}
    if (column.Width) {
      style.width = column.Width
      style.minWidth = column.Width
    }

    if (window.GLOB.columnId === column.uuid) {
      style.color = '#1890ff'
    }

    return connectDragSource(
      connectDropTarget(<th {...restProps} index={index} style={style} onDoubleClick={() => this.props.editColumn(column)}>
        <Popover overlayClassName="mk-popover-control-wrap" mouseLeaveDelay={0.2} mouseEnterDelay={0.2} content={
          <div className="mk-popover-control" onDoubleClick={(e) => e.stopPropagation()}>
            {column.type === 'colspan' ? <PlusOutlined className="plus" title="添加列" onClick={() => this.props.addElement(column)} /> : null}
            {column.type === 'custom' ? <PlusOutlined className="plus" title="添加元素" onClick={() => this.props.addElement(column)} /> : null}
            {column.type === 'custom' ? <PlusSquareOutlined className="plus" title="添加按钮" onClick={() => this.props.addElement(column, 'button')} /> : null}
            <EditOutlined className="edit" title="编辑" onClick={() => this.props.editColumn(column)} />
            <CopyOutlined title="复制显示列" style={{color: '#26C281'}} onClick={this.copycolumn} />
            {column.type === 'custom' ? <PasteComponent options={['customCardElement', 'action']} updateConfig={(res, resolve) => this.props.pasteCell(column, res, resolve)} /> : null}
            {column.type === 'custom' ? <FontColorsOutlined className="style" title="调整样式" onClick={() => this.props.changeStyle(column)}/> : null}
            <DeleteOutlined className="close" title="删除" onClick={this.deleteCol} />
            {['text', 'number', 'formula', 'extend'].includes(column.type) ? <MarkColumn field={column.field || ''} columns={fields} marks={column.marks} onSubmit={this.updateMarks} /> : null }
          </div>
        } trigger="hover">
          {children}
        </Popover>
      </th>),
    )
  }
}

const rowSource = {
  beginDrag(props) {
    return {
      index: props.index,
    }
  }
}

const ColTarget = {
  drop(props, monitor) {
    const item = monitor.getItem()
    let dragIndex = item.index
    let hoverIndex = props.index

    if (item.$init) {
      if (/sub_/.test(hoverIndex)) {
        message.warning('合并列子元素暂不支持拖拽添加。')
      } else {
        props.dropCol(item, hoverIndex.replace(new RegExp(props.componentId + '@', 'ig'), ''))
      }
      return
    } else if (dragIndex === undefined || hoverIndex === undefined || dragIndex === hoverIndex) {
      return
    } else {
      let reg = new RegExp(props.componentId + '@', 'ig')
      if (reg.test(dragIndex)) {
        props.moveCol(dragIndex.replace(reg, ''), hoverIndex.replace(reg, ''))
      } else {
        message.warning('拖动元素不在当前组件中。')
        return
      }
    }
    
    monitor.getItem().index = hoverIndex
  },
}

const DragableHeaderCol = DropTarget('col', ColTarget, connect => ({
  connectDropTarget: connect.dropTarget()
}))(
  DragSource('col', rowSource, connect => ({
    connectDragSource: connect.dragSource(),
  }))(HeaderCol),
)

class EditableColumnCell extends Component {
  updateCard = (vals, btn) => {
    const { column } = this.props
    this.props.upComponent({...column, elements: vals}, btn)
  }

  shouldComponentUpdate (nextProps, nextState) {
    const { config, column } = this.props

    if (!nextProps.column) return true

    return !is(fromJS(column), fromJS(nextProps.column)) ||
      !is(fromJS(config.columns), fromJS(nextProps.config.columns)) ||
      !is(fromJS(config.action), fromJS(nextProps.config.action))
  }

  render() {
    const { column, config, children, className, style } = this.props

    if (column && column.type === 'custom') {
      return (
        <td style={{padding: 0, ...(column.style || {})}} className={className}>
          <CardCellComponent cards={config} cardCell={column} elements={column.elements} updateElement={this.updateCard}/>
        </td>
      )
    } else if (column) {
      let val = column.field || ''
      if (column.type === 'index') {
        val = '$Index'
      } else if (column.type === 'formula') {
        val = column.formula
        if (column.eval === 'false') {
          val = val.replace(/\n/ig, '<br/>').replace(/\s/ig, '&nbsp;')
          val = <span style={{fontWeight: 'inherit'}} dangerouslySetInnerHTML={{__html: val}}></span>
        }
      }
      return (
        <td style={style} className={className}>
          {column.field ? <Paragraph className="col-copy" copyable={{ text: column.field }}>{val}</Paragraph> : val}
          {column.marks && column.marks.length ? <AntDesignOutlined className="profile"/> : null}
          {column.Hide === 'true' ? <CloseCircleOutlined className="close-col"/> : null}
        </td>
      )
    } else {
      return (
        <td style={style} className={className}>
          {children}
        </td>
      )
    }
  }
}

class NormalTableColumns extends Component {
  static propTpyes = {
    config: PropTypes.object,
    updatecolumn: PropTypes.func,
    addColumns: PropTypes.func
  }

  state = {
    tableId: '',
    data: [{uuid: Utils.getuuid()}],
    refresh: false,    // 强制刷新
    columns: [],
    fields: [],
    editStyleCard: null,
    lineMarks: []
  }

  UNSAFE_componentWillMount () {
    const { config } = this.props

    let tableId = (() => {
      let uuid = []
      let _options = 'abcdefghigklmnopqrstuv'
      for (let i = 0; i < 19; i++) {
        uuid.push(_options.substr(Math.floor(Math.random() * 0x20), 1))
      }
      return uuid.join('')
    }) ()

    this.setState({
      tableId,
      columns: fromJS(config.cols).toJS(),
      fields: fromJS(config.columns).toJS(),
      lineMarks: config.lineMarks ? fromJS(config.lineMarks).toJS() : []
    }, () => {
      const element = document.getElementById(tableId)
      element && element.style.setProperty('--mk-table-border-color', config.wrap.borderColor || '#e8e8e8')
      element && element.style.setProperty('--mk-table-color', config.wrap.color || 'rgba(0, 0, 0, 0.65)')
      element && element.style.setProperty('--mk-table-font-size', config.wrap.fontSize || '14px')
      element && element.style.setProperty('--mk-table-font-weight', config.wrap.fontWeight || 'normal')
    })
  }

  UNSAFE_componentWillReceiveProps (nextProps) {
    if (!is(fromJS(this.state.columns), fromJS(nextProps.config.cols))) {
      let _columns = fromJS(nextProps.config.cols).toJS()
      this.setState({columns: _columns})
      let lastcol = _columns.slice(-1)[0]
      if (lastcol && lastcol.focus) {
        this.editColumn(lastcol)
      }
    } else if (!is(fromJS(this.state.fields), fromJS(nextProps.config.columns))) {
      this.setState({fields: fromJS(nextProps.config.columns).toJS()})
    } else if (!is(fromJS(this.props.config.wrap), fromJS(nextProps.config.wrap))) {
      const element = document.getElementById(this.state.tableId)
      element && element.style.setProperty('--mk-table-border-color', nextProps.config.wrap.borderColor || '#e8e8e8')
      element && element.style.setProperty('--mk-table-color', nextProps.config.wrap.color || 'rgba(0, 0, 0, 0.65)')
      element && element.style.setProperty('--mk-table-font-size', nextProps.config.wrap.fontSize || '14px')
      element && element.style.setProperty('--mk-table-font-weight', nextProps.config.wrap.fontWeight || 'normal')
    }
  }

  shouldComponentUpdate (nextProps, nextState) {
    return !is(fromJS(this.state), fromJS(nextState)) || !is(fromJS(this.props.config), fromJS(nextProps.config))
  }

  moveCol = (dragIndex, hoverIndex) => {
    let _columns = fromJS(this.state.columns).toJS()

    if (/^sub_/.test(dragIndex) || /^sub_/.test(hoverIndex)) {
      let sign1 = dragIndex.split('_')
      let sign2 = hoverIndex.split('_')
      if (sign1[1] !== sign2[1]) {
        message.warning('拖动元素不在同一个合并列中。')
        return
      }
      this.loopMoveCol(_columns, sign1[1], sign1[2], sign2[2])
    } else {
      _columns.splice(hoverIndex, 0, ..._columns.splice(dragIndex, 1))
    }

    this.setState({
      columns: _columns
    }, () => {
      this.props.updatecolumn({...this.props.config, cols: _columns})
    })
  }

  loopMoveCol = (columns, colId, dragIndex, hoverIndex) => {
    columns.forEach(column => {
      if (column.type === 'colspan' && column.uuid === colId) {
        column.subcols.splice(hoverIndex, 0, ...column.subcols.splice(dragIndex, 1))
      } else if (column.type === 'colspan') {
        this.loopMoveCol(column.subcols, colId, dragIndex, hoverIndex)
      }
    })
  }

  dropCol = (item, hoverIndex) => {
    let _columns = fromJS(this.state.columns).toJS()
    let type = item.subType

    if (!['text', 'number', 'textarea', 'custom', 'formula', 'index', 'colspan'].includes(item.subType)) {
      type = 'text'
    }

    let col = { focus: true, uuid: Utils.getuuid(), Width: 120, label: 'label', field: '', type: type, elements: [] }
    if (col.type === 'colspan') {
      col.subcols = []
    } else if (col.type === 'index') {
      col.label = '序号'
    }

    _columns.splice(hoverIndex, 0, col)

    this.setState({
      columns: _columns
    }, () => {
      this.editColumn(col)
    })
  }

  loopCol = (columns, col) => {
    return columns.map(column => {
      if (column.type === 'colspan') {
        column.subcols = this.loopCol(column.subcols || [], col)
      }
      if (column.uuid === col.uuid) {
        return col
      }
      return column
    })
  }

  updateCol = (col, btn) => {
    let _columns = fromJS(this.state.columns).toJS()
    _columns = this.loopCol(_columns, col)

    this.setState({
      columns: _columns,
    }, () => {
      let config = {...this.props.config, cols: _columns}
      if (btn) {
        config.action = config.action.filter(item => item.uuid !== btn.uuid)

        setTimeout(() => {
          MKEmitter.emit('revertBtn', config.uuid)
        }, 200)
      }

      this.props.updatecolumn(config)
    })
  }

  editColumn = (col) => {
    this.setState({
      card: fromJS(col).toJS()
    })
  }

  pasteCell = (col, cell, resolve) => {
    resolve({status: true})
    
    cell.uuid = Utils.getuuid()
    cell.focus = true
    
    if (!cell.eleType) {
      if (cell.copyType === 'action') {
        cell.eleType = 'button'
        cell.width = cell.width || 12
      } else {
        cell.eleType = 'text'
      }
    }
    delete cell.copyType
    
    MKEmitter.emit('cardAddElement', col.uuid, cell)
  }

  addElement = (col, type) => {
    let column = fromJS(col).toJS()

    if (column.type === 'colspan') {
      column.subcols = column.subcols || []
      let subcol = { isSub: true, focus: true, uuid: Utils.getuuid(), label: 'label', field: '', type: 'text' }
      column.subcols.push(subcol)

      this.setState({
        card: subcol
      })
      this.updateCol(column)
    } else if (column.type === 'custom') {
      if (type === 'button') {
        let newcard = {
          uuid: Utils.getuuid(),
          focus: true,
          eleType: 'button',
          label: 'button',
          OpenType: 'prompt',
          class: 'primary',
          intertype: 'system',
          execSuccess: 'grid',
          execError: 'never',
          show: 'link'
        }
  
        // 注册事件-添加元素
        MKEmitter.emit('cardAddElement', column.uuid, newcard)
      } else {
        let newcard = {uuid: Utils.getuuid(), focus: true, width: 24, eleType: 'text', datatype: 'dynamic', style: {paddingLeft: '4px'}}
    
        // 注册事件-添加元素
        MKEmitter.emit('cardAddElement', column.uuid, newcard)
      }
    }
  }

  submitCol = (col) => {
    const { card } = this.state
    
    if (!card) return
    
    col.uuid = card.uuid
    col.isSub = card.isSub === true
    col.marks = card.marks || []

    if (col.type === 'colspan') {
      col.subcols = card.subcols || []
    } else if (col.type === 'custom') {
      col.style = card.style || {}
      col.elements = card.type === 'custom' ? (card.elements || []) : []
    }

    window.GLOB.precolumnId = window.GLOB.columnId || ''
    window.GLOB.columnId = col.uuid

    this.setState({card: null})
    this.updateCol(col)
  }

  changeStyle = (col) => {
    this.setState({
      editStyleCard: fromJS(col).toJS()
    })

    MKEmitter.emit('changeStyle', ['font', 'padding'], col.style || {}, this.getStyle)
  }

  changeSubmitStyle = () => {
    const { config } = this.props

    MKEmitter.emit('changeStyle', ['font', 'background', 'border', 'margin', 'padding'], config.submit.style || {}, this.getSubmitStyle)
  }

  getSubmitStyle = (style) => {
    const { config } = this.props

    let submit = {...config.submit, style}

    this.props.updatecolumn({...config, submit: submit})
  }

  getStyle = (style) => {
    const { editStyleCard } = this.state

    let _card = {...editStyleCard, style}
    
    this.updateCol(_card)
  }

  cancelCol = () => {
    const { card } = this.state

    if (card.focus) {
      this.deleteCol(card)
    }

    this.setState({card: null})
  }

  loopDelCol = (columns, col) => {
    return columns.filter(column => {
      if (column.type === 'colspan') {
        column.subcols = this.loopDelCol(column.subcols, col)
      }
      return column.uuid !== col.uuid
    })
  }

  deleteCol = (col) => {
    let _columns = fromJS(this.state.columns).toJS()

    if (col.type === 'colspan' && col.subcols && col.subcols.length > 0) {
      let Index = _columns.findIndex(item => item.uuid === col.uuid)

      if (Index !== -1) {
        let list = col.subcols.map(item => {
          item.isSub = false
          return item
        })

        _columns.splice(Index, 1, ...list)
      } else {
        _columns = this.loopDelCol(_columns, col)
      }
    } else {
      _columns = this.loopDelCol(_columns, col)
    }

    this.setState({
      columns: _columns
    }, () => {
      this.props.updatecolumn({...this.props.config, cols: _columns})
    })
  }

  updateLineMarks = (vals) => {
    this.setState({
      lineMarks: vals
    }, () => {
      this.props.updatecolumn({...this.props.config, lineMarks: vals})
    })
  }

  /**
   * @description 显示列复制
   */
  copycolumn = () => {
    const { columns } = this.state

    let oInput = document.createElement('input')
    let val = {
      copyType: 'editcols',
      cols: columns.filter(col => !col.origin)
    }

    let srcid = localStorage.getItem(window.location.href.split('#')[0] + 'srcId')
    if (srcid) {
      val.$srcId = srcid
    }

    oInput.value = window.btoa(window.encodeURIComponent(JSON.stringify(val)))
    document.body.appendChild(oInput)
    oInput.select()
    document.execCommand('Copy')
    oInput.className = 'oInput'
    oInput.style.display = 'none'

    message.success('复制成功。')

    document.body.removeChild(oInput)
  }

  syncfield = () => {
    const { fields } = this.state
    let columns = fromJS(this.state.columns).toJS()

    columns = columns.filter(c => !c.origin)

    let keys = columns.map(col => col.field)

    fields.forEach(item => {
      if (keys.includes(item.field)) return

      let cell = { uuid: Utils.getuuid(), label: item.label, field: item.field, Align: 'left', Hide: 'false', IsSort: 'false', Width: 120, blacklist: [], postfix: '', prefix: '', linkmenu: [], marks: [], perspective: 'linkmenu' }
      
      if (/Nvarchar|date/ig.test(item.datatype)) {
        cell.type = 'text'
        cell.rowspan = 'false'
        cell.textFormat = 'none'
      } else {
        cell.type = 'number'
        cell.format = 'none'
        cell.decimal = item.decimal || 0
        cell.Width = 80
      }

      columns.push(cell)
    })

    const _this = this

    confirm({
      content: '确定同步字段集吗？',
      onOk() {
        _this.setState({columns}, () => {
          _this.props.updatecolumn({..._this.props.config, cols: columns})
        })
      },
      onCancel() {}
    })
  }

  clear = () => {
    const _this = this

    confirm({
      content: '确定清空显示列吗？',
      onOk() {
        _this.setState({columns: []}, () => {
          _this.props.updatecolumn({..._this.props.config, cols: []})
        })
      },
      onCancel() {}
    })
  }

  verifySubmit = () => {
    const { config } = this.props
    
    this.verifyRef.handleConfirm().then(res => {
      res.style = config.submit.style || {}
      this.setState({
        visible: false
      }, () => {
        this.props.updatecolumn({...config, submit: res})
      })
    })
  }

  componentDidMount () {
    MKEmitter.addListener('submitStyle', this.getStyle)
  }

  /**
   * @description 组件销毁，清除state更新，清除快捷键设置
   */
  componentWillUnmount () {
    this.setState = () => {
      return
    }
    MKEmitter.removeListener('submitStyle', this.getStyle)
  }

  handlecolumns = (columns, fields, config, pId) => {
    return columns.map((col, index) => {
      let title = col.label
      if (col.editable === 'true') {
        title = <span>{col.label}<EditOutlined style={{position: 'absolute', bottom: 0, right: 0, color: '#1890ff', opacity: '0.7'}}/></span>
      }

      return {
        title: title,
        dataIndex: col.uuid,
        align: col.Align,
        sorter: col.IsSort === 'true',
        onCell: () => ({
          column: col,
          width: col.Width,
          config: config,
          upComponent: this.updateCol
        }),
        onHeaderCell: () => ({
          index: pId ? config.uuid + '@sub_' + pId + '_' + index : config.uuid + '@' + index,
          column: col,
          componentId: config.uuid,
          fields: fields,
          align: col.Align,
          moveCol: this.moveCol,
          dropCol: this.dropCol,
          updateCol: this.updateCol,
          addElement: this.addElement,
          editColumn: this.editColumn,
          pasteCell: this.pasteCell,
          changeStyle: this.changeStyle,
          deleteCol: this.deleteCol,
        }),
        children: col.subcols && col.subcols.length ? this.handlecolumns(col.subcols, fields, config, col.uuid) : null,
      }
    })
  }

  render() {
    const { config } = this.props
    const { fields, card, lineMarks, tableId, visible } = this.state
    const components = {
      header: {
        cell: DragableHeaderCol
      },
      body: {
        cell: EditableColumnCell
      }
    }

    let rowSelection = null
    if (config.wrap.tableType) {
      rowSelection = {
        type: config.wrap.tableType === 'radio' ? 'radio' : 'checkbox',
      }
    }

    const columns = this.handlecolumns(this.state.columns, fields, config)

    return (
      <div className={`edit-table-columns ${config.setting.laypage} ${config.wrap.mode || ''} table-vertical-${config.wrap.vertical || ''}`} id={tableId}>
        <div className="col-control">
          <Popover overlayClassName="mk-popover-control-wrap" mouseLeaveDelay={0.2} mouseEnterDelay={0.2} content={
            <div className="mk-popover-control">
              <EditOutlined className="edit" title="编辑" onClick={() => this.setState({visible: true})} />
              <FontColorsOutlined className="style" title="调整样式" onClick={this.changeSubmitStyle}/>
            </div>
          } trigger="hover">
            <Button className="submit-btn" style={config.submit.style} onDoubleClick={() => this.setState({visible: true})} type="primary">提交</Button>
          </Popover>
          <PlusOutlined title="添加列" onClick={this.props.addColumns}/>
          <CopyOutlined title="复制显示列" onClick={this.copycolumn} />
          <MarkColumn columns={fields} type="line" marks={lineMarks} onSubmit={this.updateLineMarks} />
          <ColsControl config={config} onSubmit={this.props.updatecolumn}/>
          <FileSyncOutlined title="同步字段集" onClick={this.syncfield} />
          <DeleteOutlined title="清空显示列" onClick={this.clear}/>
        </div>
        <DndProvider>
          <Table
            rowKey="uuid"
            size={config.wrap.size || 'middle'}
            bordered={config.wrap.bordered !== 'false'}
            rowSelection={rowSelection}
            components={components}
            dataSource={this.state.data}
            columns={columns}
            pagination={{
              current: 1,
              pageSize: 10,
              pageSizeOptions: ['10', '25', '50', '100', '500', '1000'],
              showSizeChanger: true,
              total: 58,
              showTotal: (total, range) => `${range[0]}-${range[1]} 共 ${total} 条`
            }}
          />
        </DndProvider>
        <EditColumn column={card} wrap={config.wrap} columns={this.state.columns} fields={fields} submitCol={this.submitCol} cancelCol={this.cancelCol}/>
        <Modal
          wrapClassName="mk-pop-modal"
          visible={visible}
          width={'90vw'}
          maskClosable={false}
          onOk={this.verifySubmit}
          onCancel={() => { this.setState({ visible: false }) }}
          destroyOnClose
        >
          <TableVerify
            config={config}
            wrappedComponentRef={(inst) => this.verifyRef = inst}
          />
        </Modal>
      </div>
    )
  }
}

export default NormalTableColumns